import numpy as np
from pdf2image import convert_from_path


def assert_pdfs_are_close(path_1: str, path_2: str, epsilon: float = 1e-10):
    """This function checks that two PDFs are close enough. We chose to convert the PDFs to images and then compare their L1 norms, because other techniques (hashing for instance) might break easily."""
    images_1 = convert_from_path(path_1)
    images_2 = convert_from_path(path_2)

    for im1, im2 in zip(images_1, images_2):
        assert np.sum(np.abs(np.array(im1) - np.array(im2))) < epsilon



"""
Test the outputs of the CLI with a bunch of different options on simple features. These tests are far from perfect.
"""

import os

import pytest

from pdf_watermark.handler import add_watermark_from_options
from pdf_watermark.options import (
    DrawingOptions,
    FilesOptions,
    GridOptions,
    InsertOptions,
)
from pdf_watermark.watermark import DEFAULTS
from tests.utils import assert_pdfs_are_close

INPUT = "tests/fixtures/input.pdf"
INPUT_UPPERCASE = "tests/fixtures/input.PDF"
OUTPUT = "output.pdf"
FIXTURES = [
    "tests/fixtures/0.pdf",
    "tests/fixtures/1.pdf",
    "tests/fixtures/2.pdf",
    "tests/fixtures/3.pdf",
    "tests/fixtures/4.pdf",
    "tests/fixtures/5.pdf",
    "tests/fixtures/6.pdf",
    "tests/fixtures/7.pdf",
]


@pytest.fixture(autouse=True)
def cleanup():
    yield
    os.remove(OUTPUT)


DRAWING_OPTIONS_FIXTURES = [
    DrawingOptions(
        watermark="watermark",
        opacity=DEFAULTS.opacity,
        angle=DEFAULTS.angle,
        text_color=DEFAULTS.text_color,
        text_font=DEFAULTS.text_font,
        text_size=DEFAULTS.text_size,
        unselectable=DEFAULTS.unselectable,
        image_scale=DEFAULTS.image_scale,
        save_as_image=DEFAULTS.save_as_image,
        dpi=DEFAULTS.dpi,
    ),
    DrawingOptions(
        watermark=r"watermark\nwith\nline\nbreaks",
        opacity=DEFAULTS.opacity,
        angle=DEFAULTS.angle,
        text_color=DEFAULTS.text_color,
        text_font=DEFAULTS.text_font,
        text_size=DEFAULTS.text_size,
        unselectable=DEFAULTS.unselectable,
        image_scale=DEFAULTS.image_scale,
        save_as_image=DEFAULTS.save_as_image,
        dpi=DEFAULTS.dpi,
    ),
]

FILES_OPTIONS_FIXTURES = [
    FilesOptions(INPUT, OUTPUT),
    FilesOptions(INPUT_UPPERCASE, OUTPUT),
]

GRID_OPTIONS_FIXTURES = [
    GridOptions(
        horizontal_boxes=DEFAULTS.horizontal_boxes,
        vertical_boxes=DEFAULTS.vertical_boxes,
        margin=DEFAULTS.margin,
    )
]

INSERT_OPTIONS_FIXTURES = [
    InsertOptions(
        y=DEFAULTS.y,
        x=DEFAULTS.x,
        horizontal_alignment=DEFAULTS.horizontal_alignment,
    )
]


def test_add_watermark_from_options():
    index = 0
    for files_options in FILES_OPTIONS_FIXTURES:
        for drawing_options in DRAWING_OPTIONS_FIXTURES:
            for specific_options in GRID_OPTIONS_FIXTURES + INSERT_OPTIONS_FIXTURES:
                add_watermark_from_options(
                    files_options=files_options,
                    drawing_options=drawing_options,
                    specific_options=specific_options,
                )
                assert_pdfs_are_close(OUTPUT, FIXTURES[index])
                index += 1



from math import cos, pi, sin
from typing import Union

import numpy as np
from reportlab.pdfgen import canvas

from pdf_watermark.options import Alignments, DrawingOptions, GridOptions, InsertOptions
from pdf_watermark.utils import (
    change_base,
    draw_centered_image,
    draw_centered_string_with_line_breaks,
    fit_image,
)


def draw_one_watermark(
    watermark: canvas.Canvas,
    x: float,
    y: float,
    rotation_matrix: np.ndarray,
    drawing_options: DrawingOptions,
    image_width: float,
    image_height: float,
):
    x_prime, y_prime = change_base(x, y, rotation_matrix)

    if drawing_options.text is not None:
        draw_centered_string_with_line_breaks(
            watermark,
            x_prime,
            y_prime,
            drawing_options.text,
        )

    if drawing_options.image is not None:
        draw_centered_image(
            watermark,
            x_prime,
            y_prime,
            image_width,
            image_height,
            drawing_options.image,
        )


def draw_insert_watermark(
    watermark: canvas.Canvas,
    drawing_options: DrawingOptions,
    specific_options: InsertOptions,
    width: float,
    height: float,
    rotation_matrix: np.ndarray,
):
    max_image_height = min(
        2 * specific_options.y * height, 2 * (2 * (1 - specific_options.y) * height)
    )
    image_width, image_height = 0, 0

    if drawing_options.text is not None:
        watermark_width = watermark.stringWidth(
            drawing_options.text, drawing_options.text_font, drawing_options.text_size
        )

    elif drawing_options.image is not None:
        if specific_options.horizontal_alignment == Alignments.LEFT.value:
            max_image_width = specific_options.x * width

        elif specific_options.horizontal_alignment == Alignments.RIGHT.value:
            max_image_width = (1 - specific_options.x) * width

        elif specific_options.horizontal_alignment == Alignments.CENTER.value:
            max_image_width = min(
                2 * (1 - specific_options.x) * width, 2 * specific_options.x * width
            )

        else:
            raise ValueError(
                f"Invalid alignment value: '{specific_options.horizontal_alignment}'."
            )

        # if the image is too big, scale it down to fit in the box
        image_width, image_height = drawing_options.image.getSize()
        image_width, image_height = fit_image(
            image_width,
            image_height,
            max_image_width,
            max_image_height,
            drawing_options.image_scale,
        )

        watermark_width = image_width

    else:
        raise ValueError("No watermark to draw.")

    if specific_options.horizontal_alignment == Alignments.LEFT.value:
        offset = -watermark_width / 2

    elif specific_options.horizontal_alignment == Alignments.RIGHT.value:
        offset = watermark_width / 2

    elif specific_options.horizontal_alignment == Alignments.CENTER.value:
        offset = 0

    else:
        raise ValueError(
            f"Invalid alignment value: '{specific_options.horizontal_alignment}'."
        )

    draw_one_watermark(
        watermark,
        specific_options.x * width + offset,
        specific_options.y * height,
        rotation_matrix,
        drawing_options,
        image_width,
        image_height,
    )


def draw_grid_watermark(
    watermark: canvas.Canvas,
    drawing_options: DrawingOptions,
    specific_options: GridOptions,
    width: float,
    height: float,
    rotation_matrix: np.ndarray,
):
    image_width, image_height = 0, 0

    horizontal_box_spacing = width / specific_options.horizontal_boxes
    vertical_box_spacing = height / specific_options.vertical_boxes

    if specific_options.margin:
        start_index = 1
    else:
        start_index = 0

    if drawing_options.image is not None:
        # if the image is too big, scale it down to fit in the box
        image_width, image_height = drawing_options.image.getSize()
        image_width, image_height = fit_image(
            image_width,
            image_height,
            horizontal_box_spacing,
            vertical_box_spacing,
            drawing_options.image_scale,
        )

    for x_index in range(start_index, specific_options.horizontal_boxes + 1):
        for y_index in range(start_index, specific_options.vertical_boxes + 1):
            # Coordinates to draw at in original coordinates system
            x_base = x_index * horizontal_box_spacing
            y_base = y_index * vertical_box_spacing

            if specific_options.margin:
                x_base -= horizontal_box_spacing / 2
                y_base -= vertical_box_spacing / 2

            draw_one_watermark(
                watermark,
                x_base,
                y_base,
                rotation_matrix,
                drawing_options,
                image_width,
                image_height,
            )


def draw_watermarks(
    file_name: str,
    width: float,
    height: float,
    drawing_options: DrawingOptions,
    specific_options: Union[GridOptions, InsertOptions],
):
    watermark = canvas.Canvas(file_name, pagesize=(width, height))

    rotation_angle_rad = drawing_options.angle * pi / 180
    rotation_matrix = np.array(
        [
            [cos(rotation_angle_rad), -sin(rotation_angle_rad)],
            [sin(rotation_angle_rad), cos(rotation_angle_rad)],
        ]
    )

    watermark.setFillColor(drawing_options.text_color, alpha=drawing_options.opacity)
    watermark.setFont(drawing_options.text_font, drawing_options.text_size)
    watermark.rotate(drawing_options.angle)

    if isinstance(specific_options, InsertOptions):
        draw_insert_watermark(
            watermark, drawing_options, specific_options, width, height, rotation_matrix
        )

    elif isinstance(specific_options, GridOptions):
        draw_grid_watermark(
            watermark, drawing_options, specific_options, width, height, rotation_matrix
        )
    else:
        raise NotImplementedError("Unknown watermark type.")

    watermark.save()



import os
from tempfile import NamedTemporaryFile
from typing import Union

import pypdf

from pdf_watermark.draw import draw_watermarks
from pdf_watermark.options import (
    DrawingOptions,
    FilesOptions,
    GridOptions,
    InsertOptions,
)
from pdf_watermark.utils import convert_content_to_images, sort_pages


def add_watermark_to_pdf(
    input: str,
    output: str,
    drawing_options: DrawingOptions,
    specific_options: Union[GridOptions, InsertOptions],
):
    pdf_writer = pypdf.PdfWriter()
    pdf_to_transform = pypdf.PdfReader(input)

    page_sizes = []
    for page in pdf_to_transform.pages:
        page_sizes.append((page.mediabox.width, page.mediabox.height))

    order = []
    # Only one watermark is computed per page size
    for watermark_width, watermark_height in set(page_sizes):
        with NamedTemporaryFile(delete=False) as temporary_file:
            # The watermark is stored in a temporary pdf file
            draw_watermarks(
                temporary_file.name,
                watermark_width,
                watermark_height,
                drawing_options,
                specific_options,
            )

            if drawing_options.unselectable and not drawing_options.save_as_image:
                convert_content_to_images(
                    temporary_file.name,
                    drawing_options.dpi,
                )

            watermark_pdf = pypdf.PdfReader(temporary_file.name)

            # Add watermark to pages with the same size
            for index, (page, (page_width, page_height)) in enumerate(
                zip(pdf_to_transform.pages, page_sizes)
            ):
                if page_width == watermark_width and page_height == watermark_height:
                    page.merge_page(watermark_pdf.pages[0])
                    pdf_writer.add_page(page)
                    order.append(index)

            # Remove temp file - https://stackoverflow.com/questions/23212435/permission-denied-to  -write-to-my-temporary-file
            temporary_file.close()
            os.unlink(temporary_file.name)

    pdf_writer = sort_pages(pdf_writer, order)

    with open(output, "wb") as f:
        pdf_writer.write(f)

    if drawing_options.save_as_image:
        convert_content_to_images(output, drawing_options.dpi)


def add_watermark_from_options(
    files_options: FilesOptions,
    drawing_options: DrawingOptions,
    specific_options: Union[GridOptions, InsertOptions],
):
    for input_file, output_file in files_options:
        add_watermark_to_pdf(input_file, output_file, drawing_options, specific_options)



"""
Test for PDFs with different page sizes.
"""

import os

import pytest

from pdf_watermark.handler import add_watermark_from_options
from pdf_watermark.options import DrawingOptions, FilesOptions, GridOptions
from pdf_watermark.watermark import DEFAULTS
from tests.utils import assert_pdfs_are_close

INPUT = "tests/fixtures/different_sizes_input.pdf"
OUTPUT = "output.pdf"
FIXTURE = "tests/fixtures/different_sizes_output.pdf"


@pytest.fixture(autouse=True)
def cleanup():
    yield
    os.remove(OUTPUT)


DRAWING_OPTIONS_FIXTURE = DrawingOptions(
    watermark="watermark",
    opacity=DEFAULTS.opacity,
    angle=DEFAULTS.angle,
    text_color=DEFAULTS.text_color,
    text_font=DEFAULTS.text_font,
    text_size=DEFAULTS.text_size,
    unselectable=DEFAULTS.unselectable,
    image_scale=DEFAULTS.image_scale,
    save_as_image=DEFAULTS.save_as_image,
    dpi=DEFAULTS.dpi,
)

FILES_OPTIONS_FIXTURE = FilesOptions(INPUT, OUTPUT)

GRID_OPTIONS_FIXTURE = GridOptions(
    horizontal_boxes=DEFAULTS.horizontal_boxes,
    vertical_boxes=DEFAULTS.vertical_boxes,
    margin=DEFAULTS.margin,
)


def test_different_page_sizes():
    add_watermark_from_options(
        files_options=FILES_OPTIONS_FIXTURE,
        drawing_options=DRAWING_OPTIONS_FIXTURE,
        specific_options=GRID_OPTIONS_FIXTURE,
    )
    assert_pdfs_are_close(OUTPUT, FIXTURE)



from functools import wraps

import click

from pdf_watermark.handler import add_watermark_from_options
from pdf_watermark.options import (
    DrawingOptions,
    FilesOptions,
    GridOptions,
    InsertOptions,
)


class DEFAULTS:
    angle = 45
    dpi = 300
    horizontal_alignment = "center"
    horizontal_boxes = 3
    image_scale = 1
    margin = False
    opacity = 0.1
    save_as_image = False
    text_color = "#000000"
    text_font = "Helvetica"
    text_size = 12
    unselectable = False
    vertical_boxes = 6
    x = 0.5
    y = 0.5


def generic_watermark_parameters(f):
    @wraps(f)
    @click.argument("file")
    @click.argument("watermark")
    @click.option(
        "-s",
        "--save",
        type=str,
        help="File or folder to save results to. By default, the input files are overwritten.",
    )
    @click.option(
        "-o",
        "--opacity",
        type=float,
        help="Watermark opacity between 0 (invisible) and 1 (no transparency).",
        default=DEFAULTS.opacity,
    )
    @click.option(
        "-a",
        "--angle",
        type=float,
        help="Watermark inclination in degrees.",
        default=DEFAULTS.angle,
    )
    @click.option(
        "-tc",
        "--text-color",
        type=str,
        help="Text color in hexadecimal format, e.g. #000000.",
        default=DEFAULTS.text_color,
    )
    @click.option(
        "-tf",
        "--text-font",
        type=str,
        help="Text font to use. Supported fonts are those supported by reportlab.",
        default=DEFAULTS.text_font,
    )
    @click.option(
        "-ts",
        "--text-size",
        type=int,
        help="Text font size.",
        default=DEFAULTS.text_size,
    )
    @click.option(
        "--unselectable",
        type=bool,
        is_flag=True,
        help="Make the watermark text unselectable. This works by drawing the text as an image, and thus results in a larger file size.",
        default=DEFAULTS.unselectable,
    )
    @click.option(
        "-is",
        "--image-scale",
        type=float,
        help="Scale factor for the image. Note that before this factor is applied, the image is already scaled down to fit in the boxes.",
        default=DEFAULTS.image_scale,
    )
    @click.option(
        "--save-as-image",
        type=bool,
        is_flag=True,
        help="Convert each PDF page to an image. This makes removing the watermark more difficult but also increases the file size.",
        default=DEFAULTS.save_as_image,
    )
    @click.option(
        "--dpi",
        type=int,
        help="DPI to use when saving the PDF as an image.",
        default=DEFAULTS.dpi,
    )
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


@click.group()
def cli():
    """
    Add a watermark to one or more PDF files.

    The watermark can be repeated in a grid pattern using the grid command, or inserted at a specific position using the insert command.
    """
    pass


@cli.command()
@click.option(
    "-y",
    "--y",
    type=float,
    help="Position of the watermark with respect to the vertical direction. Must be between 0 and 1.",
    default=DEFAULTS.y,
)
@click.option(
    "-x",
    "--x",
    type=float,
    help="Position of the watermark with respect to the horizontal direction. Must be between 0 and 1.",
    default=DEFAULTS.x,
)
@click.option(
    "-ha",
    "--horizontal-alignment",
    type=str,
    help="Alignment of the watermark with respect to the horizontal direction. Can be one of 'left', 'right' and 'center'.",
    default=DEFAULTS.horizontal_alignment,
)
@generic_watermark_parameters
def insert(
    file,
    watermark,
    save,
    opacity,
    angle,
    text_color,
    text_font,
    text_size,
    unselectable,
    image_scale,
    save_as_image,
    dpi,
    y,
    x,
    horizontal_alignment,
):
    """
    Add a watermark at a specific position.

    Add a WATERMARK to one or more PDF files referenced by FILE.
    WATERMARK can be either a string or a path to an image file.
    FILE can be a single file or a directory, in which case all PDF files in the directory will be watermarked.
    """
    add_watermark_from_options(
        FilesOptions(input=file, output=save),
        DrawingOptions(
            watermark=watermark,
            opacity=opacity,
            angle=angle,
            text_color=text_color,
            text_font=text_font,
            text_size=text_size,
            unselectable=unselectable,
            image_scale=image_scale,
            save_as_image=save_as_image,
            dpi=dpi,
        ),
        InsertOptions(
            y=y,
            x=x,
            horizontal_alignment=horizontal_alignment,
        ),
    )


@cli.command()
@click.option(
    "-h",
    "--horizontal-boxes",
    type=int,
    help="Number of repetitions of the watermark along the horizontal direction.",
    default=DEFAULTS.horizontal_boxes,
)
@click.option(
    "-v",
    "--vertical-boxes",
    type=int,
    help="Number of repetitions of the watermark along the vertical direction.",
    default=DEFAULTS.vertical_boxes,
)
@click.option(
    "-m",
    "--margin",
    type=bool,
    is_flag=True,
    help="Wether to leave a margin around the page or not. When False (default), the watermark will be cut on the PDF edges.",
    default=DEFAULTS.margin,
)
@generic_watermark_parameters
def grid(
    file,
    watermark,
    save,
    opacity,
    angle,
    text_color,
    text_font,
    text_size,
    unselectable,
    image_scale,
    save_as_image,
    dpi,
    horizontal_boxes,
    vertical_boxes,
    margin,
):
    """
    Add a watermark in a grid pattern.

    Add a WATERMARK to one or more PDF files referenced by FILE.
    WATERMARK can be either a string or a path to an image file.
    FILE can be a single file or a directory, in which case all PDF files in the directory will be watermarked.
    """
    add_watermark_from_options(
        FilesOptions(input=file, output=save),
        DrawingOptions(
            watermark=watermark,
            opacity=opacity,
            angle=angle,
            text_color=text_color,
            text_font=text_font,
            text_size=text_size,
            unselectable=unselectable,
            image_scale=image_scale,
            save_as_image=save_as_image,
            dpi=dpi,
        ),
        GridOptions(
            horizontal_boxes=horizontal_boxes,
            vertical_boxes=vertical_boxes,
            margin=margin,
        ),
    )


if __name__ == "__main__":
    cli()



import os
from enum import Enum
from typing import List, Union

from reportlab.lib.colors import HexColor
from reportlab.lib.utils import ImageReader


class DrawingOptions:
    def __init__(
        self,
        watermark: str,
        opacity: float,
        angle: float,
        text_color: str,
        text_font: str,
        text_size: int,
        unselectable: bool,
        image_scale: float,
        save_as_image: bool,
        dpi: int,
    ) -> None:
        self.image = None
        self.text = None

        potential_image_path = os.path.join(os.getcwd(), watermark)
        if watermark.endswith(
            (".png", ".jpg", ".jpeg", ".PNG", ".JPG", ".JPEG")
        ) and os.path.isfile(potential_image_path):
            self.image = ImageReader(watermark)
        else:
            self.text = watermark

        self.opacity = opacity
        self.angle = angle
        self.text_color = HexColor(text_color)
        self.text_font = text_font
        self.text_size = text_size
        self.unselectable = unselectable
        self.image_scale = image_scale
        self.save_as_image = save_as_image
        self.dpi = dpi


class FilesOptions:
    def __init__(
        self,
        input: str,
        output: Union[None, str],
    ) -> None:
        input = os.path.join(os.getcwd(), input)

        if not os.path.exists(input):
            raise ValueError("Input file or directory does not exist.")
        elif os.path.isdir(input):
            if output is not None and output.endswith((".pdf", ".PDF")):
                raise ValueError(
                    "Output must be a directory when input is a directory."
                )
        elif os.path.isfile(input) and input.endswith((".pdf", ".PDF")):
            if output is not None and not output.endswith((".pdf", ".PDF")):
                raise ValueError("Output must be a pdf file when input is a pdf file.")
        else:
            raise ValueError("Input must be a pdf file or a directory.")

        if output is None:
            output = input
        else:
            output = os.path.join(os.getcwd(), output)

        if os.path.isfile(input):
            self.input_files = [input]
            if output is None:
                self.output_files = [input]
            else:
                self.output_files = [output]
        else:
            self.input_files = []
            self.output_files = []
            self.add_directory_to_files(input, output)

    def add_directory_to_files(
        self, input_directory: str, output_directory: str
    ) -> None:
        if not os.path.isdir(input_directory):
            raise ValueError("Directory argument must be a directory.")

        for file in os.listdir(input_directory):
            input_path = os.path.join(input_directory, file)
            output_path = os.path.join(output_directory, file)

            if os.path.isdir(input_path):
                self.add_directory_to_files(input_path, output_path)

            elif os.path.isfile(input_path) and input_path.endswith((".pdf", ".PDF")):
                self.input_files.append(input_path)
                self.output_files.append(output_path)

                if not os.path.exists(output_directory):
                    os.makedirs(output_directory)

    def __iter__(self):
        return iter(zip(self.input_files, self.output_files))

    def __next__(self):
        return next(zip(self.input_files, self.output_files))


class GridOptions:
    def __init__(
        self,
        horizontal_boxes: int,
        vertical_boxes: int,
        margin: bool,
    ) -> None:
        self.horizontal_boxes = horizontal_boxes
        self.vertical_boxes = vertical_boxes
        self.margin = margin


class Alignments(Enum):
    LEFT = "left"
    RIGHT = "right"
    CENTER = "center"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class InsertOptions:
    def __init__(
        self,
        y: float,
        x: float,
        horizontal_alignment: List[Alignments],
    ) -> None:
        if not Alignments.has_value(horizontal_alignment):
            raise Exception(
                "Invalid argument. Horizontal alignment must be either left, right or center."
            )
        self.y = y
        self.x = x
        self.horizontal_alignment = horizontal_alignment



from io import BytesIO
from typing import List, Tuple

import numpy as np
import pypdf
from pdf2image import convert_from_path
from pdf2image.exceptions import PopplerNotInstalledError
from reportlab.lib.utils import ImageReader
from reportlab.pdfgen import canvas


def draw_centered_image(
    canvas: canvas.Canvas,
    x: float,
    y: float,
    width: float,
    height: float,
    image: ImageReader,
):
    bottom_left_x = x - width / 2
    bottom_left_y = y - height / 2
    canvas.drawImage(
        image,
        bottom_left_x,
        bottom_left_y,
        width=width,
        height=height,
        mask="auto",
    )


def draw_centered_string_with_line_breaks(
    watermark: canvas.Canvas,
    x: float,
    y: float,
    text: str,
):
    text_lines = text.split(r"\n")
    line_height = (
        watermark._leading
    )  # line height is set when setting the font of the canvas
    y += (len(text_lines) - 1) * line_height / 2  # also center the text vertically
    for line in text_lines:
        watermark.drawCentredString(x, y, line)
        y -= watermark._leading


def change_base(x: float, y: float, rotation_matrix: np.ndarray) -> Tuple[float, float]:
    # Since we rotated the original coordinates system, use the inverse of the rotation matrix
    # (which is the transposed matrix) to get the coordinates we have to draw at
    new_coordinates = np.transpose(rotation_matrix) @ np.array([[x], [y]])
    return new_coordinates[0, 0], new_coordinates[1, 0]


def fit_image(image_width, image_height, max_image_width, max_image_height, scale):
    if image_width > max_image_width:
        change_ratio = max_image_width / image_width
        image_width = max_image_width
        image_height *= change_ratio
    if image_height > max_image_height:
        change_ratio = max_image_height / image_height
        image_height = max_image_height
        image_width *= change_ratio

    image_width *= scale
    image_height *= scale

    return image_width, image_height


def convert_content_to_images(file_name: str, dpi: int):
    # load pages as images
    try:
        images = convert_from_path(file_name, dpi=dpi, fmt="png", transparent=True)
    except PopplerNotInstalledError:
        print(
            "Warning : the --save-as-image and --unselectable options require poppler to be installed. Proceeding without these options. Pleaser refer to the documentation for more information."
        )
        return None

    # get the page sizes
    pdf_to_transform = pypdf.PdfReader(file_name)
    page_sizes = []
    for page in pdf_to_transform.pages:
        page_sizes.append((page.mediabox.width, page.mediabox.height))

    # create new pdf
    pdf = canvas.Canvas(file_name)

    for image, (page_width, page_height) in zip(images, page_sizes):
        pdf.setPageSize((page_width, page_height))
        compressed = BytesIO()
        image.save(compressed, format="png", optimize=True, quality=dpi // 10)

        pdf.drawImage(
            ImageReader(compressed),
            0,
            0,
            width=page_width,
            height=page_height,
            mask="auto",
        )
        pdf.showPage()

    pdf.save()


def sort_pages(pdf: pypdf.PdfWriter, order: List[int]):
    output = pypdf.PdfWriter()
    for index in np.argsort(order):
        output.add_page(pdf.pages[int(index)])

    return output
